
/*
 * Copyright (c) 2008 - 2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and proprietary
 * rights in and to this software, related documentation and any modifications thereto.
 * Any use, reproduction, disclosure or distribution of this software and related
 * documentation without an express license agreement from NVIDIA Corporation is strictly
 * prohibited.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS*
 * AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS OR IMPLIED,
 * INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT
 * LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
 * BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES
 */

#include <optix.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sutil.h>
#include <ImageLoader.h>
#include <sampleConfig.h>
#include <optixu/optixpp.h>
#include <GLUTDisplay.h>
#include <sutilCommonDefines.h>
#include <Mouse.h>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <optixu/optixu_aabb.h>

#undef SAMPLES_DIR
#define SAMPLES_DIR "C:\\NVIDIA_OptiX_SDK_1.0.0\\SDK\\tutorial"
//#define SAMPLES_DIR ".\\"


using namespace optixu;

static unsigned int WIDTH = 640;
static unsigned int HEIGHT = 480;

enum AnimState {
  ANIM_NONE,      // pause all
  ANIM_PHASE,     // phase animation
  ANIM_PHASE2,    // phase animation with locked phase1
  ANIM_ALL,       // full auto
};
static AnimState animstate = ANIM_ALL;

// Animated camera.
struct AnimCamera
{
  AnimCamera()
    : m_pos( make_float3(0) )
    , m_aspect( (float)WIDTH/(float)HEIGHT )
    , m_t( 0 )
  {}

  void update( double t )
  {
    m_t += t * 0.1;
    m_pos.y = (float)( 2 + sin( m_t*1.5 ) );
    m_pos.x = (float)( 2.3*sin( m_t ) );
    m_pos.z = (float)( 0.5+2.1*cos( m_t ) );
  }

  void apply( Context context )
  {
    PinholeCamera pc( m_pos, make_float3(0), make_float3(0,1,0), 60.f, 60.f/m_aspect );
    float3 eye, u, v, w;
    pc.getEyeUVW( eye, u, v, w );
    context["eye"]->setFloat( eye );
    context["U"]->setFloat( u );
    context["V"]->setFloat( v );
    context["W"]->setFloat( w );
  }

  float3 m_pos;
  float  m_aspect;
  double m_t;
};

class MandelbulbScene : public SampleScene
{
public:

  MandelbulbScene( const std::string& texture_path )
    : m_alpha( 0.5f ),        // scaled down with HEIGHT to adapt to screen resolution
      m_delta( 1.0f ),        // exponent for distance that goes into computation of epsilon value
      m_max_iterations( 5 ),  // fractal formula iteration level
      m_power( 8 ),           // power of fractal, defaults to 8
      m_epsilon( 1e-7f ),     // epsilon lower bound
      m_factor( 1.0f ),       // raymarching factor for distance stepping
      m_tempo( 1.0 ),        // animation tempo
      m_threshold( 1000.0f ), // threshold for deciding whether point diverges
      m_radius( 2.0f ),       // radius. This large sphere radius allows us to move the camera
                              // very close (and even inside) without provoking overstepping.
      m_phase1( 0.0f ),       // phase1 offset for angle phi
      m_phase( 0.0f ),        // phase offset for angle phi
      m_exagerate ( 1.0f ),   // distance exageration (used to enable "go inside" feature)
      m_color_t( 0 ),         // color setting
      texture_path( texture_path )
  {}

  // From SampleScene
  virtual void   initScene( InitialCameraData& camera_data );
  virtual void   trace( const RayGenCameraData& camera_data );
  virtual void   cleanUp();
  virtual Buffer getOutputBuffer();
  virtual bool   keyPressed(unsigned char key, int x, int y);

  virtual void resize(unsigned int width, unsigned int height);

  std::string texpath( const std::string& base );
  void createGeometry();

  static bool m_useGLBuffer;

private:

  Context context;

  // parameters of Equation 16 of [Hart et al., 1989]
  float  m_alpha;
  float  m_delta;

  float m_epsilon;
  float m_factor;
  float m_threshold;
  float m_tempo;
  float m_radius;
  float m_phase1;
  float m_phase;
  float m_exagerate;

  unsigned int m_max_iterations;
  int m_power;
  unsigned int m_color_t;

  AnimCamera  m_cam;        // camera animation

  GeometryGroup m_geometrygroup;

  std::string   texture_path;
};

// make PBOs default since it's a lot of faster
bool         MandelbulbScene::m_useGLBuffer = true;

void MandelbulbScene::resize(unsigned int width, unsigned int height)
{
  Buffer buffer = getOutputBuffer();
  SampleScene::resize(width, height);
  m_cam.m_aspect = (float)width / (float)height;

  WIDTH=width;
  HEIGHT=height;
  context[ "alpha" ]->setFloat( m_alpha/HEIGHT );
}


void MandelbulbScene::initScene( InitialCameraData& camera_data )
{
  try {
    // Setup state
    context = Context::create();
    context->setRayTypeCount( 2 );
    context->setEntryPointCount( 1 );
    context->setStackSize(1425);

    context["max_depth"]->setInt( 1 );             // maximum reflection depth
    context["radiance_ray_type"]->setUint( 0u );   // not currently used in .cu file
    context["shadow_ray_type"]->setUint( 1u );     // not currently used in .cu file
    context["scene_epsilon"]->setFloat( 1.e-4f );
    context["color_t"]->setFloat( 0.0f );

    Variable output_buffer = context["output_buffer"];

    output_buffer->set(createOutputBuffer(context, RT_FORMAT_UNSIGNED_BYTE4, WIDTH, HEIGHT));

    // Set up camera
    camera_data = InitialCameraData( make_float3( 0.0f,+3.0f,  0.0f ), // eye
                                     make_float3( 0.0f, 0.0f,  0.0f ), // lookat
                                     make_float3( 0.0f, 0.0f,  1.0f ), // up
                                     60.0f );                          // vfov

    // Declare camera variables.  The values do not matter, they will be overwritten in trace.
    context["eye"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    context["U"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    context["V"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );
    context["W"]->setFloat( make_float3( 0.0f, 0.0f, 0.0f ) );

    // Ray gen program
    Program ray_gen_program = context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "pinhole_camera" );
    context->setRayGenerationProgram( 0, ray_gen_program );

    // Exception
    Program exception_program = context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "exception" );
    context->setExceptionProgram( 0, exception_program );
    context["bad_color"]->setFloat( 1.0f, 1.0f, 0.0f );

    // Miss program
    context->setMissProgram( 0, context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "envmap_miss" ) );
    const float3 default_color = make_float3(1.0f, 1.0f, 1.0f);
    context["envmap"]->setTextureSampler( loadTexture( context, texpath("CedarCity.hdr"), default_color) );
    context["bg_color"]->setFloat( make_float3(108.0f/255.0f, 166.0f/255.0f, 205.0f/255.0f) * 0.5f );

    // Setup lights
    context["ambient_light_color"]->setFloat(0.5f,0.4f,0.3f);
    BasicLight lights[] = { 
      { { -50.0f, 100.0f, 5.0f }, { 1.0f, 0.8f, 0.6f }, 1 },
    };

    Buffer light_buffer = context->createBuffer(RT_BUFFER_INPUT);
    light_buffer->setFormat(RT_FORMAT_USER);
    light_buffer->setElementSize(sizeof(BasicLight));
    light_buffer->setSize( sizeof(lights)/sizeof(lights[0]) );
    memcpy(light_buffer->map(), lights, sizeof(lights));
    light_buffer->unmap();

    context["lights"]->set(light_buffer);

    // Create scene geom
    createGeometry();

    // Finalize
    context->validate();
    context->compile();

  } catch( Exception& e ){
    sutilReportError( e.getErrorString().c_str() );
    exit(1);
  }
}


Buffer MandelbulbScene::getOutputBuffer()
{
  return context["output_buffer"]->getBuffer();
}


void MandelbulbScene::trace( const RayGenCameraData& camera_data )
{
  static bool last_cont = GLUTDisplay::getContinuousDrawingFlag();
  const bool cont = GLUTDisplay::getContinuousDrawingFlag();

  static double last_t = 0;
  if( cont && !last_cont )  // check if continuous drawing was just turned on
    last_t = 0;
  double t;
  sutilCurrentTime( &t );
  const double time_elapsed = last_t > 0 ? t-last_t : 0;

  last_t = t;
  last_cont = cont;

  if( animstate == ANIM_PHASE )
  {
    // phase offset for angle phi
    m_phase1 = m_phase = fmodf((float)(m_phase + (2 * M_PI / 30) * m_tempo * time_elapsed), (float)(2*M_PI));
    context[ "phase1" ]->setFloat( m_phase1 );
    context[ "phase" ]->setFloat( m_phase );
  }
  if( animstate == ANIM_PHASE2 )
  {
    // phase offset for angle phi
    m_phase = fmodf((float)(m_phase + (2 * M_PI / 30) * m_tempo * time_elapsed), (float)(2*M_PI));
    m_phase1 = 0.0f;
    context[ "phase1" ]->setFloat( m_phase1 );
    context[ "phase" ]->setFloat( m_phase );
  }

  // update camera
  if( animstate == ANIM_ALL )
  {
    m_cam.update( m_tempo * time_elapsed );
    m_cam.apply( context );

    const float f = log(10.0f+1.0f)/log(2.0f); // proper scaling for sine values
    static double m_time = 180.0 * asin(1.0 / f) / M_PI;  // start with a power 1 "ball"
    m_time += m_tempo * time_elapsed;
    // sine-sweep between a power 10 fractal and a power -10 fractal
    float m_pwr = (float) (f * sin(M_PI*m_time/180.0));
    // use an exponential growth law for the power to make it more interesting
    float rendered_power = m_pwr > 0 ? -1.0f+powf(2.0f, fabs(m_pwr)) : 1.0f-powf(2.0f, fabs(m_pwr));
    context[ "power" ]->setFloat( rendered_power );

    m_phase1 = m_phase = 0.0f;
    context[ "phase1" ]->setFloat( m_phase1 );
    context[ "phase" ]->setFloat( m_phase );
  }
  else
  {
    context["eye"]->setFloat( camera_data.eye );
    context["U"]->setFloat( camera_data.U );
    context["V"]->setFloat( camera_data.V );
    context["W"]->setFloat( camera_data.W );
  }

  // trace
  Buffer buffer = context["output_buffer"]->getBuffer();
  RTsize buffer_width, buffer_height;
  buffer->getSize( buffer_width, buffer_height );
  context->launch( 0, 
                   static_cast<unsigned int>(buffer_width),
                   static_cast<unsigned int>(buffer_height)
                   );
}


void MandelbulbScene::cleanUp()
{
  context->destroy();
}


std::string MandelbulbScene::texpath( const std::string& base )
{
  return texture_path + PATH_SEPARATOR + base;
}

void MandelbulbScene::createGeometry()
{
  // Mandelbulb object
  Geometry mandelbulb = context->createGeometry();
  mandelbulb->setPrimitiveCount( 1u );
  mandelbulb->setBoundingBoxProgram( context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "bounds" ) );
  mandelbulb->setIntersectionProgram( context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "intersect" ) );
  
  Program mandelbulb_ch = context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "mandelbulb_ch_radiance" );
  Program mandelbulb_ah = context->createProgramFromPTXFile( ptxpath( "mandelbulb.cu" ), "mandelbulb_ah_shadow" );

  // Mandelbulb material
  Material mandelbulb_matl = context->createMaterial();
  mandelbulb_matl->setClosestHitProgram( 0, mandelbulb_ch );
  mandelbulb_matl->setAnyHitProgram( 1, mandelbulb_ah );

  // Set material properties for phong shading
  context["Ka"]->setFloat(.25f, 0.2f, 0.025f);// ambient color
  context["Kd"]->setFloat(1.0f, 0.8f, 0.1f);  // direct lighting color
  context["Ks"]->setFloat(1.0f, 1.0f, 1.0f);  // specular highlight color
  context["phong_exp"]->setFloat(32);         // exponent for phong highlights
  context["reflectivity"]->setFloat(.5f, 0.4f, 0.05f); // reflectivity per color component (this is also modulated with specular)

  // Place geometry into hierarchy
  std::vector<GeometryInstance> gis;
  gis.push_back( context->createGeometryInstance( mandelbulb, &mandelbulb_matl, &mandelbulb_matl+1 ) );

  m_geometrygroup = context->createGeometryGroup();
  m_geometrygroup->setChildCount( static_cast<unsigned int>(gis.size()) );
  for(size_t i = 0; i < gis.size(); ++i) {
    m_geometrygroup->setChild( (int)i, gis[i] );
  }
  m_geometrygroup->setAcceleration( context->createAcceleration("NoAccel","NoAccel") );


  // Top level group
  Group topgroup = context->createGroup();
  topgroup->setChildCount( 1 );
  topgroup->setChild( 0, m_geometrygroup );
  topgroup->setAcceleration( context->createAcceleration("Bvh","Bvh") );

  context["top_object"]->set( topgroup );
  context["top_shadower"]->set( m_geometrygroup );

  // set QJS parameters
  context[ "alpha" ]->setFloat( m_alpha/HEIGHT );
  context[ "delta" ]->setFloat( m_delta );

  // number of iterations for recursive fractal formula
  context[ "max_iterations" ]->setUint( m_max_iterations );
  // power of fractal formula
  context[ "power" ]->setFloat( (float)m_power );
  // epsilon for iteration termination
  context[ "eps" ]->setFloat( m_epsilon );
  // raymarching factor for distance stepping
  context[ "fact" ]->setFloat( m_factor );
  // threshold for deciding whether point diverges
  context[ "threshold" ]->setFloat( m_threshold );
  // radius
  context[ "radius" ]->setFloat( m_radius );
  // phase1 offset for angle phi
  context[ "phase1" ]->setFloat( m_phase1 );
  // phase offset for angle phi
  context[ "phase" ]->setFloat( m_phase );
  // distance exageration (used to enable "go inside" feature)
  context[ "exag" ]->setFloat( m_exagerate );
  // ambient occlusion switch defaults to on
  context[ "occswitch" ]->setUint( 1 );
  // phong highlight switch defaults to on
  context[ "phongswitch" ]->setUint( 1 );
  // shadow switch defaults to off
  context[ "shadowswitch" ]->setUint( 0 );
   // reflection switch defaults to off
  context[ "refswitch" ]->setUint( 0 );
   // background color switch defaults to on
  context[ "bgswitch" ]->setUint( 1 );
}

bool MandelbulbScene::keyPressed(unsigned char key, int x, int y)
{
   switch (key)
   {
     case 'd':
       m_alpha *= pow(0.1f, 1.0f/10.0f);
       std::cerr << "alpha: " << m_alpha/HEIGHT << std::endl;
       context[ "alpha" ]->setFloat( m_alpha/HEIGHT );
       return true;
     case 'D':
       m_alpha *= pow(10.0f, 1.0f/10.0f);
       std::cerr << "alpha: " << m_alpha/HEIGHT << std::endl;
       context[ "alpha" ]->setFloat( m_alpha/HEIGHT );
       return true;
     case 'I':
       m_max_iterations = std::max(m_max_iterations - 1u, 1u);
       std::cerr << "max_iterations: " << m_max_iterations << std::endl;
       context[ "max_iterations" ]->setUint( m_max_iterations );
       return true;
     case 'i':
       m_max_iterations++;
       std::cerr << "max_iterations: " << m_max_iterations << std::endl;
       context[ "max_iterations" ]->setUint( m_max_iterations );
       return true;
     case 'g':
         if (m_exagerate == 1.0f) {
           m_exagerate = 3.0f;
           std::cerr << "go-inside mode on. Try moving the camera into the bulb." << std::endl;
         } else {
           m_exagerate = 1.0f;
           std::cerr << "go-inside mode off. Moving the camera inside is now impossible." << std::endl;
         }
//       std::cerr << "exagerate: " << m_exagerate << std::endl;
       context[ "exag" ]->setFloat( m_exagerate );
       return true;
     case 'S':
       m_factor *= pow(10.0f, 1.0f/10.0f);
       std::cerr << "factor: " << m_factor << std::endl;
       context[ "fact" ]->setFloat( m_factor );
       return true;
     case 's':
       m_factor *= pow(0.1f, 1.0f/10.0f);
       std::cerr << "factor: " << m_factor << std::endl;
       context[ "fact" ]->setFloat( m_factor );
       return true;
//     case 'T':
//       m_threshold *= pow(10.0f, 1.0f/10.0f);
//       std::cerr << "threshold: " << m_threshold << std::endl;
//       context[ "threshold" ]->setFloat( m_threshold );
//       return true;
//     case 't':
//       m_threshold *= pow(0.1f, 1.0f/10.0f);
//       std::cerr << "threshold: " << m_threshold << std::endl;
//       context[ "threshold" ]->setFloat( m_threshold );
//       return true;
     case 't':
       m_tempo *= pow(0.5f, 1.0f/4.0f);
       std::cerr << "tempo: " << m_tempo << std::endl;
       return true;
     case 'T':
       m_tempo *= pow(2.0f, 1.0f/4.0f);
       std::cerr << "tempo: " << m_tempo << std::endl;
       return true;
//     case 'Y':
//       m_radius *= pow(2.0f, 1.0f/10.0f);
//       std::cerr << "radius: " << m_radius << std::endl;
//       context[ "radius" ]->setFloat( m_radius );
//       return true;
//     case 'y':
//       m_radius *= pow(0.5f, 1.0f/10.0f);
//       std::cerr << "radius: " << m_radius << std::endl;
//       context[ "radius" ]->setFloat( m_radius );
//       return true;
     case '-':
       m_power--;
       std::cerr << "power: " << m_power << std::endl;
       context[ "power" ]->setFloat( (float)m_power );
       return true;
     case '+':
       m_power++;
       std::cerr << "power: " << m_power << std::endl;
       context[ "power" ]->setFloat( (float)m_power );
       return true;
     case ' ': {
       animstate = AnimState( (animstate+1) % 4 );
       if (animstate != ANIM_ALL) {
           context[ "power" ]->setFloat( (float)m_power );
           context["color_t"]->setFloat( (float)(m_color_t) );
       }
       return true;
     case 'a':
       context[ "occswitch" ]->setUint( !!! context[ "occswitch" ]->getUint() );
       return true;
     case 'r':
       context[ "refswitch" ]->setUint( !!! context[ "refswitch" ]->getUint() );
       return true;
     case 'p':
       context[ "phongswitch" ]->setUint( !!! context[ "phongswitch" ]->getUint() );
       return true;
     case 'o':
       context[ "shadowswitch" ]->setUint( !!! context[ "shadowswitch" ]->getUint() );
       return true;
     case 'b':
       context[ "bgswitch" ]->setUint( !!! context[ "bgswitch" ]->getUint() );
       return true;
     case 'm':
       return true;
//     case 'c':
//       // update mandelbulb set color
//       m_color_t = (m_color_t+1)%3;
//       context["color_t"]->setFloat( (float)(m_color_t) );
//       return true;
      }
   }
   return false;
}

void printHelp()
{
  std::cerr <<
    "NVIDIA OptiX interactive ray tracing example\n"
    "Usage options for Mandelbulb:\n"
    "\n"
    "Runtime Argument Options:  \n"
    "  -h or -?      Print this usage message\n"
    "  -s            Save each frame to file frame_XXXXX.ppm in current folder\n"
    "  -d=<dim>   Initial window dimensions in form: =640x480  \n"
    "\n"
    "\n"
    "Interactive options:\n"
    "  spacebar             Cycles between 3 animation modes, interactive mode\n"
    "  left mouse           Camera Rotate/Orbit (when interactive)\n"
    "  middle mouse         Camera Pan/Truck (when interactive)\n"
    "  right mouse          Camera Dolly (when interactive)\n"
    "  right mouse + shift  Camera FOV (when interactive)\n"
    "  f                    Full screen mode toggle\n"
    "  q or ESC             Quits the example\n"
    "  t/T                  Adjusts animation tempo in animated mode\n"
    "  -/+                  Decreases or increases power of the fractal formula\n"
    "  i/I                  Change iteration limit for fractal formula\n"
    "  d/D                  Adjusts detail level (epsilon bound computation)\n"
    "  s/S                  Adjusts step factor for raymarching\n"
//    "  t/T                  Adjusts divergence threshold\n"
//    "  y/Y                  Adjusts bounding sphere radius\n"
    "  a                    Toggle ambient occlusion term\n"
    "  p                    Toggle phong shading highlights\n"
    "  r                    Toggle reflections on/off\n"
    "  b                    Toggle background image\n"
    "  g                    Toggle go-inside mode\n"
    ;
}

void printUsageAndExit( const std::string& argv0 )
{
#ifdef RELEASE
  printHelp();
#else
  std::cerr << "Usage  : " << argv0 << " [options]\n"
            << "Options: --help           | -h | -?        Print this usage message\n"
            << "         --pbo            | -p             Use OpenGL PBO for output buffer (default)\n"
            << "         --nopbo          | -n             Use internal output buffer\n"
            << std::endl;
  GLUTDisplay::printUsage();
#endif
  exit(1);
}

int main( int argc, char** argv )
{
  GLUTDisplay::init( argc, argv );

  std::string texture_path;

  for(int i = 1; i < argc; ++i) {
    std::string arg = argv[i];
    if(arg == "-p" || arg == "--pbo") {
      MandelbulbScene::m_useGLBuffer = true;
    } else if( arg == "-n" || arg == "--nopbo" ) {
      MandelbulbScene::m_useGLBuffer = false;
    } else if ( arg == "-t" || arg == "--texture-path" ) {
      if ( i == argc-1 ) {
        printUsageAndExit( argv[0] );
      }
      texture_path = argv[++i];
    } else if( arg == "-?" || arg == "-h" || arg == "--help" ) {
      printUsageAndExit(argv[0]);
    } else {
      std::cerr << "Unknown option '" << arg << "'\n";
      printUsageAndExit(argv[0]);
    }
  }

  printHelp();

  if( texture_path.empty() ) {
    texture_path = std::string( SAMPLES_DIR ) + PATH_SEPARATOR + "data";
  }

  try {
    MandelbulbScene scene( texture_path );
    GLUTDisplay::run( "Mandelbulb - hit space bar for interactive mode, f for fullscreen mode", &scene, true );
  } catch( Exception& e ){
    sutilReportError( e.getErrorString().c_str() );
    exit(1);
  }

  return 0;
}
