
/*
 * Copyright (c) 2008 - 2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and proprietary
 * rights in and to this software, related documentation and any modifications thereto.
 * Any use, reproduction, disclosure or distribution of this software and related
 * documentation without an express license agreement from NVIDIA Corporation is strictly
 * prohibited.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED *AS IS*
 * AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS OR IMPLIED,
 * INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS BE LIABLE FOR ANY
 * SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES WHATSOEVER (INCLUDING, WITHOUT
 * LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
 * BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGES
 */

#include <optix_cuda.h>
#include <optix_math.h>
#include <optixu/optixu_matrix.h>
#include <optixu/optixu_aabb.h>
#include <sutilCommonDefines.h>
#include "helpers.h"



// References:
// [1] Hart, J. C., Sandin, D. J., and Kauffman, L. H. 1989. Ray tracing deterministic 3D fractals
// [2] http://www.devmaster.net/forums/showthread.php?t=4448


// Parameters for Mandelbulb ray intersection code

rtDeclareVariable( float,  alpha , , );
rtDeclareVariable( float,  delta , , );

rtDeclareVariable( float,  color_t , , );           // 0,1,2 are full colors, in between is morph
rtDeclareVariable( uint,   max_iterations , , );    // max iterations for divergence determination
rtDeclareVariable( float,  power , , );             // power of fractal
rtDeclareVariable( float,  eps , , );               // epsilon for termination of iteration
rtDeclareVariable( float,  fact , , );              // raymarching factor for distance stepping
rtDeclareVariable( float,  threshold , , );         // threshold for deciding whether point diverges to infinity
rtDeclareVariable( float,  radius , , );            // bounding sphere radius
rtDeclareVariable( float,  phase1 , , );            // initial phase offset for angle phi
rtDeclareVariable( float,  phase , , );             // phase offset for angle phi
rtDeclareVariable( float,  exag , , );              // distance exageration (used to enable "go inside" feature)


// Parameters for Mandelbulb shading code

// Scene global stuff
rtDeclareVariable(int,      max_depth, , );
rtDeclareVariable(float,    scene_epsilon, , );
rtDeclareVariable(float3,   ambient_light_color, , );
rtDeclareVariable(float3,   bg_color, , );

// Phong shading and lights
rtDeclareVariable(float3,   Ka, , );
rtDeclareVariable(float3,   Kd, , );
rtDeclareVariable(float3,   Ks, , );
rtDeclareVariable(float3,   reflectivity, , );
rtDeclareVariable(float,    phong_exp, , );
rtBuffer<BasicLight> lights;

// switches
rtDeclareVariable( uint,   occswitch , , );         // ambient occlusion switch
rtDeclareVariable( uint,   shadowswitch , , );      // shadow ray switch
rtDeclareVariable( uint,   phongswitch , , );       // phong shading switch
rtDeclareVariable( uint,   refswitch , , );         // reflection switch
rtDeclareVariable( uint,   bgswitch , , );          // background image switch


// Mandelbulb set intersection code object output attributes

rtDeclareVariable(float3, normal, attribute normal, ); 
rtDeclareVariable(float,  orbit,  attribute orbit, ); 
rtDeclareVariable(float,  outeps, attribute outeps, ); 


// Ray variables

rtDeclareVariable(optix::Ray, ray, rtCurrentRay, );

struct PerRayData_radiance
{
  float3 result;
  int depth;
  float epsilon;
};

struct PerRayData_shadow
{
  float3 attenuation;
  float epsilon;
};

rtDeclareVariable(rtObject, top_object, , );
rtDeclareVariable(rtObject, top_shadower, , );
rtDeclareVariable(float, isect_t, rtIntersectionDistance, );
rtDeclareVariable(PerRayData_radiance, prd_radiance, rtPayload, );
rtDeclareVariable(PerRayData_shadow,   prd_shadow,   rtPayload, );


// Camera variables

rtDeclareVariable(float3,        eye, , );
rtDeclareVariable(float3,        U, , );
rtDeclareVariable(float3,        V, , );
rtDeclareVariable(float3,        W, , );
rtDeclareVariable(float3,        bad_color, , );
rtBuffer<uchar4, 2>              output_buffer;
rtDeclareVariable(unsigned int,  radiance_ray_type, , );

rtDeclareVariable(uint2, launch_index, rtLaunchIndex, );
rtDeclareVariable(float, time_view_scale, , ) = 1e-6f;


//
// Pinhole Camera
//

//#define TIME_VIEW

RT_PROGRAM void pinhole_camera()
{
#ifdef TIME_VIEW
  clock_t t0 = clock(); 
#endif
  uint2 screen = output_buffer.size();

  float2 d = make_float2(launch_index) / make_float2(screen) * 2.f - 1.f;
  float3 ray_origin = eye;
  float3 ray_direction = normalize(d.x*U + d.y*V + W);
  
  optix::Ray ray = optix::make_Ray(ray_origin, ray_direction, radiance_ray_type, scene_epsilon, RT_DEFAULT_MAX);

  PerRayData_radiance prd;
  prd.depth = 0;
  prd.epsilon = 0.0f;

  rtTrace(top_object, ray, prd);

#ifdef TIME_VIEW
  clock_t t1 = clock(); 
 
  float expected_fps   = 1.0f;
  float pixel_time     = ( t1 - t0 ) * time_view_scale * expected_fps;
  output_buffer[launch_index] = make_color( make_float3(  pixel_time ) ); 
#else
  output_buffer[launch_index] = make_color( prd.result );
#endif
}

RT_PROGRAM void exception()
{
  output_buffer[launch_index] = make_color( bad_color );
}


//
// Environment map background
//
rtTextureSampler<float4, 2> envmap;
RT_PROGRAM void envmap_miss()
{
  if (bgswitch)
  {
    float theta = atan2f( ray.direction.x, ray.direction.y );
    float phi   = M_PIf * 0.5f -  acosf( ray.direction.z );
    float u     = (theta + M_PIf) * (0.5f * M_1_PIf);
    float v     = 0.5f * ( 1.0f + sin(phi) );
    prd_radiance.result = make_float3( tex2D(envmap, u, v) );
  }
  else
  {
    prd_radiance.result = bg_color;
  }
}


//
// Intersect the bounding sphere of the Mandelbulb set.
//
__host__ __device__ bool intersectBoundingSphere( float3 o, float3 d, float& tmin, float &tmax )
{
  const float sq_radius = radius * radius;
  const float b = dot( o, d );
  const float c = dot( o, o ) - sq_radius;
  const float disc = b*b - c;
  
  if( disc > 0.0f )
  {
    const float sdisc = sqrtf( disc );
    tmin = (-b - sdisc);
    tmax = (-b + sdisc);

    if(tmin > tmax)
    {
      const float temp = tmin;
      tmax = tmin;
      tmin = temp;
    }

    return true;
  } else {
    tmin = tmax = 0;
  }
  return false;
}


//
// Distance estimation class for Mandelbulbs
//
struct MandelbulbSet
{
  __host__ __device__
  MandelbulbSet(const unsigned int max_iterations):m_max_iterations(max_iterations) {
      m_min_dst = 999999.f;
  }

  // Return the lowest orbit radius found during iteration
  __device__ float get_orbit() const
  {
      return m_min_dst;
  }


  // Return the approximate lower bound on the distance from x to the set.
  __device__
  float operator()( float3 z )
  {
    const float3 c = z;                    // constant term in iterative formula
    const float p = power;                 // power of fractal

    // divergence threshold nearly infinite for negative powers
    const float d_threshold = (p < 0) ? 1E9 : min(2.0f, pow(threshold,1.0f/p));

    // orbit trapping continues at existing m_min_dist
    float min_dst = m_min_dst;

    // point z polar coordinates
    float R  = length(z);
    float th = atan2(z.y, z.x);
    float ph = asinf(z.z / R) + phase1;  // initial phase offset

    // orbit trapping relative to point (0,0,0)
    if (R < min_dst) min_dst = R;

    float3 dz;
    float ph_dz = 0.0f;
    float th_dz = 0.0f;
    float R_dz  = 1.0f;

    // Iterate to compute the distance estimator.
    int i = m_max_iterations;

    while( i-- )
    {
      if (p>0)
      {
        // positive powers p

        // purely scalar dz iteration (Thanks to Enforcer for the great tip)
        float zr = powf(R, p-1);
        R_dz = zr*R_dz*p + 1;
        
        // z iteration
        float P_ = zr * R; // equivalent to __powf(R, p);
        float s1, c1; __sincosf(p*ph, &s1, &c1);
        float s2, c2; __sincosf(p*th, &s2, &c2);
        z.x = P_ * c1*c2 + c.x;
        z.y = P_ * c1*s2 + c.y;
        z.z = P_ * s1    + c.z;
      }
      else
      {
        // negative powers p

        // derivative dz iteration
        float pP_ = p*__powf(R, p-1.0f);
        float c1, s1;   __sincosf(ph_dz+(p-1.0f)*ph, &s1, &c1);
        float s1_, c1_; __sincosf(th_dz+(p-1.0f)*th, &s1_, &c1_);
        dz.x = pP_ * R_dz*c1*c1_+1.0f;
        dz.y = pP_ * R_dz*c1*s1_;
        dz.z = pP_ * R_dz*s1;
          
        // polar coordinates of derivative dz
        R_dz  = sqrtf(dz.x*dz.x + dz.y*dz.y + dz.z*dz.z);
        th_dz = atan2(dz.y, dz.x);
        ph_dz = asinf(dz.z /R_dz);

        // z iteration
        float P_ = __powf(R, p);
        float s2, c2; __sincosf(p*ph, &s2, &c2);
        float s2_, c2_; __sincosf(p*th, &s2_, &c2_);
        z.x = P_ * c2*c2_ + c.x;
        z.y = P_ * c2*s2_ + c.y;
        z.z = P_ * s2    + c.z;
      }
      // compute new length of z
      R  = length(z);

      // results are not stored for the "extra" iteration at i == 0
      // orbit trapping relative to point (0,0,0)
      if (R < min_dst) min_dst = R;

      // Stop when we know the point diverges and return the result.
      if( R > d_threshold ) break;

      // compute remaining polar coordinates of z and iterate further
      th = atan2(z.y, z.x);
      ph = asinf(z.z / R) + phase;  // iterative phase offset
    }

    // return the result if we reached convergence
    m_min_dst = min_dst;

    return 0.5 * R * logf(R)/R_dz;
  }

  unsigned int m_max_iterations;
  float m_min_dst;
};


//
// Gradient utility function
//
template<typename SignedDistance>
__device__
float3 grad(SignedDistance d, const float3 &x, const float eps = 1e-6f)
{
  float dx = d(x + make_float3(eps,    0,   0)) - d(x - make_float3(eps,   0,   0));
  float dy = d(x + make_float3(  0,  eps,   0)) - d(x - make_float3(  0, eps,   0));
  float dz = d(x + make_float3(  0,    0, eps)) - d(x - make_float3(  0,   0, eps));

  float f = 1.0f/(2*eps);
  return f * make_float3(dx, dy, dz);
}


//
// Normal utility function
//
template<typename SignedDistance>
__device__
float3 estimate_normal(const SignedDistance distance,
                       const float3 &x,
                       const float epsilon = 1e-3f)
{
  return normalize(grad(distance, x, epsilon));
}


//
// Ray marching against a Distance Estimator
//
RT_PROGRAM void intersect(int primIdx)
{
  float tmin, tmax;
  if( intersectBoundingSphere(ray.origin, ray.direction, tmin, tmax) )
  {
    MandelbulbSet distance( max_iterations );

    float3 ray_direction = ray.direction;
    float3 x = ray.origin + tmin * ray_direction;
    const float eps_ = eps;
    // if the ray has previously collided with the fractal, the ray will have received an
    // associated epsilon value which is valid for the hit point. We continue intersecting
    // with the fractal using the same epsilon value.
    const float ray_epsilon = ray.ray_type == 0 ? prd_radiance.epsilon : prd_shadow.epsilon;
    float epsilon = ray_epsilon;
    const float exagerate = exag;  // distance exageration factor.
                                   // To be honest I have no idea why exagerating the
                                   // intersection point with the ray by a certain factor
                                   // allows the camera to go "inside" the bulb but I am
                                   // not questioning it because it is a cool effect.
                                   // however it breaks shadows and causes weird self-
                                   // reflections

    float dist_from_origin = tmin;
    float dist = distance( x );
    bool found = false;
    float factor = fact;

     // limit factor to <= 0.25 for negative power to minimize overstepping artifacts
    if (power < 0) factor = min(fact, 0.25);

    // number of raymarching steps scales inversely with factor
    for( unsigned int i = 0; i < (int)(500.f / factor); ++i )
    {
      // check if by stepping forward we would end up "inside" the set
      float ndist;
      if ((ndist = distance( x + factor * dist * ray_direction )) < 0)
      {
        // if so, postulate we have found the surface and bail out
        found = true; break;
      }

      // step forward on the ray by dist * factor
      x += factor * dist * ray_direction;
      dist_from_origin += exagerate*factor * dist;  // include possible exageration here

      // distance-dependent computation of epsilon. this is far from perfect.
      if (ray_epsilon == 0.0f)
          epsilon = max(eps_, alpha * powf(dist_from_origin/exagerate, delta));

      // left the bounding sphere?
      if (dist_from_origin > tmin + exagerate*(tmax-tmin)) break;

      // Check if distance is within epsilon or if we marched too far.
      if( (dist >= 0 && dist < epsilon) )
      {
        // did we reach the surface (within epsilon)?
        found = true;
        break;
      }

      // the previously obtained distance from above becomes the next distance to step forward
      dist = ndist;
    }

    // Found intersection?
    if( found )
    {
      if( rtPotentialIntersection( dist_from_origin ) )
      {
        outeps = epsilon;
        orbit  = distance.get_orbit();
        normal = estimate_normal(distance, x, epsilon/2);

        rtReportIntersection( 0 );
      }
    }
  }
}


//
// Bounding box for aabb algorithm
//
RT_PROGRAM void bounds (int, float result[6])
{
  optix::Aabb* aabb = (optix::Aabb*)result;
  const float sz = radius;
  aabb->m_min = make_float3(-sz);
  aabb->m_max = make_float3(sz);
}


//
// Phong shader for Mandelbulb set.
//
RT_PROGRAM void mandelbulb_ch_radiance()
{
  // Phong lighting
  float3 hit_point = ray.origin + isect_t * ray.direction;

  float3 ffnormal = faceforward( normal, -ray.direction, normal );

  // optional ambient occlusion (occlusion term defined by orbit trap)
  float occlusion = 1.0f;
  if (occswitch)
  {
    occlusion = orbit * orbit;
    occlusion = clamp(occlusion, 0.2f, 1.0f);
  }

  // Ambient component
  float3 ambient = Ka * ambient_light_color;

  float3 direct = make_float3( 0.0f );
  for(int i = 0; i < lights.size(); ++i) {
    BasicLight light = lights[i];
    float3 L  = normalize(light.pos - hit_point);
    float nDl = dot( ffnormal, L);
    if( nDl > 0.0f ) {

      PerRayData_shadow shadow_prd;
      shadow_prd.attenuation = make_float3(1.0f);
      if (shadowswitch && light.casts_shadow && exag == 1.0f)
      {
        // cast shadow ray
        float Ldist = length(light.pos - hit_point);
        optix::Ray shadow_ray( hit_point + outeps * normal, L, 1 /* any_hit ray type */, scene_epsilon, Ldist );
        shadow_prd.epsilon = outeps;
        rtTrace(top_shadower, shadow_ray, shadow_prd);
      }

      if( fmaxf(shadow_prd.attenuation) > 0.0f ) { 

        // Lambertian
        float3 diffuse = Kd * nDl * light.color;

        float3 specular = make_float3( 0.0f );
        if (phongswitch)
        {
          // Phong highlight
          float3 H = normalize(L - ray.direction);
          float nDh = dot( ffnormal, H );
          if( nDh > 0.0f ) // shadow attenuation goes into specular component squared
            specular = Ks * shadow_prd.attenuation * light.color * powf( nDh, phong_exp );
        }

        direct += shadow_prd.attenuation * (diffuse + specular);
      }
    }
  }

  // reflection ray
  float3 reflected = make_float3( 0.0f );
  if (refswitch)
  {
    if( prd_radiance.depth < max_depth) {
      PerRayData_radiance refl_prd;
      refl_prd.depth = prd_radiance.depth+1;
      float3 R = reflect( ray.direction, ffnormal );
      optix::Ray refl_ray( hit_point + outeps * normal, R, 0, scene_epsilon, RT_DEFAULT_MAX );
      refl_prd.epsilon = outeps;
      rtTrace(top_object, refl_ray, refl_prd);
      reflected = reflectivity * Ks * refl_prd.result;
    }
  }

  prd_radiance.result = (ambient + direct + reflected) * occlusion;
}


//
// Shadow hit program for Mandelbulb
//
RT_PROGRAM void mandelbulb_ah_shadow()
{
  // this material is opaque, so it fully attenuates all shadow rays
  prd_shadow.attenuation = make_float3(0);
  rtTerminateRay();
}
